# ---------------------------------------------------------------------
# Gufo Err: TracebackMiddleware
# ---------------------------------------------------------------------
# Copyright (C) 2022, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
from typing import Callable, Iterable, Tuple
from pprint import pformat

# Gufo Labs modules
from ..abc.middleware import BaseMiddleware
from ..types import ErrorInfo, FrameInfo
from ..logger import logger


class TracebackMiddleware(BaseMiddleware):
    """
    Dump traceback to the `gufo.err` logger.

    Args:
        format: dumping format, one of `terse`, `extend`.

    Examples:
        Implicit initialization of the middleware using
        default `terse` format:

        ``` py
        from gufo.err import err

        err.setup()
        ```

        Implicit initialization of the middleware using
        explicit `terse` format:

        ``` py
        from gufo.err import err

        err.setup(format="terse")
        ```

        Implicit initialization of the middleware using
        explicit `extend` format:

        ``` py
        from gufo.err import err

        err.setup(format="extend")
        ```

        Explicit initialization of the middleware:

        ``` py
        from gufo.err import err
        from gufo.err.middleware.traceback import TracebackMiddleware

        err.setup(middleware=[TracebackMiddleware(format="extend")])
        ```
    """

    SEP = "-" * 79
    MAX_VAR_LEN = 72

    def __init__(self, format: str = "terse") -> None:
        super().__init__()
        try:
            self.format: Callable[[ErrorInfo], Iterable[str]] = getattr(
                self, f"iter_fmt_{format}"
            )
        except AttributeError:
            raise ValueError(f"Invalid format {format}")

    def process(self, info: ErrorInfo) -> None:
        msg = "\n".join(self.format(info))
        logger.error(msg)

    def iter_stack(self, err: ErrorInfo) -> Iterable[FrameInfo]:
        """
        Iterate stack according to direction.

        Args:
            err: ErrorInfo instance.

        Returns:
            Iterable of FrameInfo
        """
        yield from err.stack

    def traceback_message(self) -> str:
        """
        Get proper traceback message.

        Returns:
            String like "Traceback (most resent call last):"
        """
        return "Traceback (most resent call last):"

    def iter_vars(self, fi: FrameInfo) -> Iterable[Tuple[str, str]]:
        """
        Iterate frame variables and convert them to the readable form.

        Args:
            fi: FrameInfo instance

        Returns:
            Iterable of (`var name`, `var value`).
        """
        for k, v in fi.locals.items():
            try:
                rv = repr(v)
                if len(rv) > self.MAX_VAR_LEN:
                    rv = pformat(v)
            except Exception as e:
                rv = f"repr() failed: {e}"
            yield k, rv

    def iter_fmt_terse(self, err: ErrorInfo) -> Iterable[str]:
        yield f"Error: {err.fingerprint}"
        yield self.traceback_message()
        for fi in self.iter_stack(err):
            if fi.source:
                yield (
                    f'  File "{fi.source.file_name}", '
                    f"line {fi.source.current_line}, in {fi.name}"
                )
                line = fi.source.lines[
                    fi.source.current_line - fi.source.first_line
                ].lstrip()
                yield f"    {line}"
            else:
                yield '  File "<stdin>", line ??? in <module>'
        yield f"{err.exception.__class__.__name__}: {str(err.exception)}"

    def iter_fmt_extend(self, err: ErrorInfo) -> Iterable[str]:
        yield f"Error: {err.fingerprint}"
        yield f"{err.exception.__class__.__name__}: {str(err.exception)}"
        yield self.traceback_message()
        for fi in self.iter_stack(err):
            yield self.SEP
            if fi.source:
                yield (
                    f"File: {fi.source.file_name} "
                    f"(line {fi.source.current_line})"
                )
                for n, line in enumerate(
                    fi.source.lines, start=fi.source.first_line
                ):
                    sign = "==>" if n == fi.source.current_line else "   "
                    yield f"{n:5d} {sign} {line}"
            else:
                yield "File: <stdin> (line ???)"
            if fi.locals:
                yield "Locals:"
                for var_name, var_value in self.iter_vars(fi):
                    if len(var_value) > self.MAX_VAR_LEN:
                        yield f"{var_name:>20s} |\n{var_value}"
                    else:
                        yield f"{var_name:>20s} = {var_value}"
        yield self.SEP



# ---------------------------------------------------------------------
# Gufo Err: TypesFailFast
# ---------------------------------------------------------------------
# Copyright (C) 2022, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
from typing import Type, Iterable
from types import TracebackType

# Gufo Labs modules
from ..abc.failfast import BaseFailFast


class TypesFailFast(BaseFailFast):
    """
    Fail-fast on the given list of exception types.

    Args:
        types: Iterable of exception types.

    Example:

        ```
        from gufo.err import err
        from gufo.err.types import TypesFailFast

        err.setup(fail_fast=[TypesFailFast([RuntimeError, ValueError])])
        ```
    """

    def __init__(self, types: Iterable[Type[Exception]]) -> None:
        super().__init__()
        self.types = set(types)

    def must_die(
        self,
        t: Type[BaseException],
        v: BaseException,
        tb: TracebackType,
    ) -> bool:
        return t in self.types



# ---------------------------------------------------------------------
# Gufo Err: err singleton
# ---------------------------------------------------------------------
# Copyright (C) 2022, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
import sys
import os
from typing import Optional, Type, List, Iterable
from types import TracebackType
from uuid import UUID
import hashlib

# Gufo Labs modules
from .types import ErrorInfo, FrameInfo
from .frame import iter_frames
from .abc.failfast import BaseFailFast
from .abc.middleware import BaseMiddleware
from .logger import logger


DEFAULT_NAME = "unknown"
DEFAULT_VERSION = "unknown"
DEFAULT_HASH = "sha1"
DEFAULT_EXIT_CODE = 1


class Err(object):
    """
    Error handling singleton.

    Example:

        ``` py
        from gufo.err import err

        err.setup()
        ```
    """

    def __init__(self) -> None:
        self.__name = DEFAULT_NAME
        self.__version = DEFAULT_VERSION
        self.__hash_fn = hashlib.sha1
        self.__initialized = False
        self.__failfast_chain: List[BaseFailFast] = []
        self.__middleware_chain: List[BaseMiddleware] = []
        self.__failfast_code = DEFAULT_EXIT_CODE
        self.__root_module: Optional[str] = None

    def process(self) -> None:
        """
        Process current exception context in the fenced code block.

        Example:

            ``` py
            from gufo.err import err

            ...
            try:
                my_function()
            except Exception:
                err.process()
            ```
        """
        t, v, tb = sys.exc_info()
        if not t or not v or not tb:
            return  # Not an exception context
        self.__process(t, v, tb)

    def __process(
        self,
        t: Type[BaseException],
        v: BaseException,
        tb: Optional[TracebackType] = None,
    ) -> None:
        """
        Process given exception context. Called either from .process()
        or as sys.excepthook for unhandled exceptions.

        Args:
            t: Exception type.
            v: Exception value.
            tb: Traceback frame.
        """
        if not self.__initialized:
            raise RuntimeError("setup() is not called")
        if not tb:
            return
        if t in (SystemExit, KeyboardInterrupt):
            raise  # Do not mess the exit sequence
        if self.__must_die(t, v, tb):
            os._exit(self.__failfast_code)  # Fatal error, die quickly
        # Collect stack frames
        # @todo: separate handling of endless recursion
        stack = list(iter_frames(tb))
        # Calculate error fingerprint
        fp = self.__fingerprint(t, v, stack)
        # Build stack info
        err_info = ErrorInfo(
            name=self.__name,
            version=self.__version,
            fingerprint=fp,
            stack=stack,
            exception=v,
        )
        # Process the response
        self.__run_middleware(err_info)

    def setup(
        self,
        *,
        catch_all: bool = False,
        root_module: Optional[str] = None,
        name: str = DEFAULT_NAME,
        version: str = DEFAULT_VERSION,
        hash: str = DEFAULT_HASH,
        fail_fast: Optional[Iterable[BaseFailFast]] = None,
        fail_fast_code: int = DEFAULT_EXIT_CODE,
        middleware: Optional[Iterable[BaseMiddleware]] = None,
        format: Optional[str] = "terse",
        error_info_path: Optional[str] = None,
        error_info_compress: Optional[str] = None,
    ) -> "Err":
        """
        Setup error handling singleton. Must be called
        only once. Raises RuntimeError when called twice.

        Args:
            catch_all: Install global system exception hook.
            name: Application or service name.
            version: Application or service version.
            root_module: Top-level application module/namespace for
                split stack fingerprinting. Topmost frame from the root or
                the nested modules will be considered in the error
                fingerprint.
            hash: Fingerprint hashing function name. Available functions
                are: sha256, sha3_512, blake2s, sha3_224, md5, sha384,
                sha3_256, shake_256, blake2b, sha224, shake_128, sha3_384,
                sha1, sha512. Refer to the Python's hashlib for details.
            fail_fast: Iterable of BaseFailFast instances for fail-fast
                detection.
                Process will terminate with `fail_fast_code` error code
                if any of instances in the chain will return True.
            fail_fast_code: System exit code on fail-fast termination.
            middleware: Iterable of BaseMiddleware instances
                for error processing middleware.
                Instances are evaluated in the order of appearance.
            format: If not None install TracebackMiddleware for given
                output format.
            error_info_path: If not None install ErrorInfoMiddleware.
                `error_info_path` should point to a writable directories,
                in which the error info files to be written.
            error_info_compress: Used only with `error_info_path`.
                Set error info compression method. One of:

                * `None` - do not compress
                * `gz` - GZip
                * `bz2` - BZip2
                * `xz` - LZMA/xz

        Returns:
            Err instance.
        """
        if self.__initialized:
            raise RuntimeError("Already initialized")
        # Install system-wide exception hook
        if catch_all:
            sys.excepthook = self.__process
        # Init parameters
        self.__name = name or DEFAULT_NAME
        self.__version = version or DEFAULT_VERSION
        self.__failfast_code = fail_fast_code
        self.__root_module = root_module
        try:
            self.__hash_fn = getattr(hashlib, hash)
        except AttributeError:
            raise RuntimeError(f"Unknown hash: {hash}")
        # Initialize fail fast chain
        if fail_fast:
            self.__failfast_chain = []
            for ff in fail_fast:
                self.add_fail_fast(ff)
        else:
            self.__failfast_chain = []
        # Initialize response chain
        if middleware:
            self.__middleware_chain = self.__default_middleware(
                format=format,
                error_info_path=error_info_path,
                error_info_compress=error_info_compress,
            )
            for resp in middleware:
                self.add_middleware(resp)
        else:
            self.__middleware_chain = self.__default_middleware(
                format=format,
                error_info_path=error_info_path,
                error_info_compress=error_info_compress,
            )
        # Mark as initialized
        self.__initialized = True
        return self

    def __must_die(
        self,
        t: Type[BaseException],
        v: BaseException,
        tb: TracebackType,
    ) -> bool:
        """
        Process fail-fast sequence and return True if the
        process must die quickly.
        """
        if not tb:
            return False
        return any(ff.must_die(t, v, tb) for ff in self.__failfast_chain)

    def __run_middleware(self, err_info: ErrorInfo) -> None:
        """
        Process all the middleware.

        Args:
            err_info: Filled ErrorInfo structure
        """
        for resp in self.__middleware_chain:
            try:
                resp.process(err_info)
            except Exception as e:
                logger.error("%r middleware failed: %s", resp, e)

    def iter_fingerprint_parts(
        self,
        t: Type[BaseException],
        v: BaseException,
        stack: List[FrameInfo],
    ) -> Iterable[str]:
        """
        Iterable to yield all fingerprint parts.
        May be overriden in subclasses.

        Args:
            t: Exception type.
            v: Exception instance:
            stack: Current stack.
        Returns:
            Iterable of strings.
        """
        yield self.__name  # Service name
        yield self.__version  # Service version
        yield t.__name__  # Exception class
        # Top-level stack info
        if stack:
            top = stack[0]
            yield top.module or "unknown"  # Top module
            yield top.name  # Top callable name
            if top.source:
                yield str(top.source.current_line)  # Top execution line
        # Application stack info
        if self.__root_module:
            app_top = None
            prefix = f"{self.__root_module}."
            for frame in stack:
                if frame.module and (
                    frame.module == self.__root_module
                    or frame.module.startswith(prefix)
                ):
                    app_top = frame
                    break
            if app_top:
                yield app_top.module or "unknown"  # App module
                yield app_top.name  # App module Current callable name
                if app_top.source:
                    yield str(
                        app_top.source.current_line
                    )  # App execution line

    def __fingerprint(
        self,
        t: Type[BaseException],
        v: BaseException,
        stack: List[FrameInfo],
    ) -> UUID:
        """
        Calculate error fingerprint for given exception
        and the stack. Fingerprint is stable for repeating
        conditions.

        Args:
            t: Exception type.
            v: Exception instance:
            stack: Current stack.

        Returns:
            Error fingerprint as UUID.
        """
        fp_hash = self.__hash_fn(
            b"\x00".join(
                x.encode("utf-8")
                for x in self.iter_fingerprint_parts(t, v, stack)
            )
        ).digest()
        return UUID(bytes=fp_hash[:16], version=5)

    def add_fail_fast(self, ff: BaseFailFast) -> None:
        """
        Add fail-fast handler to the end of the chain.

        Args:
            ff: BaseFailFast instance.
        """
        if not isinstance(ff, BaseFailFast):
            raise ValueError(
                "add_fail_fast() argument must be BaseFailFast instance"
            )
        self.__failfast_chain.append(ff)

    def add_middleware(self, mw: BaseMiddleware) -> None:
        """
        Add middleware to the end of the chain.

        Args:
            mw: BaseMiddleware instance
        """
        if not isinstance(mw, BaseMiddleware):
            raise ValueError(
                "add_response() argument must be BaseResponse instance"
            )
        self.__middleware_chain.append(mw)

    def __default_middleware(
        self,
        format: Optional[str] = None,
        error_info_path: Optional[str] = None,
        error_info_compress: Optional[str] = None,
    ) -> List[BaseMiddleware]:
        """
        Get default middleware chain.

        Args:
            format: traceback format. See TracebackMiddleware for details.
                Do not configure tracebacks if None.
            error_info_path: Directory path to write error info.
                See ErrorInfoMiddleware for details.
                Do not configure middleware if None.
            error_info_compress: Error info compression algorithm. Used along
                with `error_info_path`.
        """
        r: List[BaseMiddleware] = []
        if format is not None:
            from .middleware.traceback import TracebackMiddleware

            r.append(TracebackMiddleware(format=format))
        if error_info_path is not None:
            from .middleware.errorinfo import ErrorInfoMiddleware

            r.append(
                ErrorInfoMiddleware(
                    path=error_info_path, compress=error_info_compress
                )
            )
        return r


# Define the singleton
err = Err()



# ---------------------------------------------------------------------
# Gufo Err: Frame Extraction
# ---------------------------------------------------------------------
# Copyright (C) 2022, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
import sys
from types import TracebackType
from typing import Optional, Iterable, cast
from importlib.abc import InspectLoader

# Gufo Labs modules
from .types import FrameInfo, SourceInfo


def exc_traceback() -> TracebackType:
    """
    Extract and return top-level excecution frame
    from current exception context.

    Returns:
        Top-level exception frame.
    """
    return cast(TracebackType, sys.exc_info()[2])


def iter_frames(
    tb: TracebackType, context_lines: int = 7
) -> Iterable[FrameInfo]:
    """
    Args:
        tb: current execution frame.
        context_lines: Source code context to extract.
            Current line, up to `context_lines` below
            the current line, and up to `context_lines`
            above the current line will be extracted.

    Returns:
        Iterable of FrameInfo, starting from top of the
        stack (current code position).
    """
    current: Optional[TracebackType] = tb
    while current is not None:
        frame = current.tb_frame
        src = __get_lines(
            file_name=frame.f_code.co_filename,
            line_no=current.tb_lineno,
            context_lines=context_lines,
            loader=frame.f_globals.get("__loader__"),
            module_name=frame.f_globals.get("__name__"),
        )
        yield FrameInfo(
            name=frame.f_code.co_name,
            module=frame.f_globals.get("__name__"),
            source=src,
            locals=current.tb_frame.f_locals,
        )
        current = current.tb_next


def __source_from_loader(
    loader: InspectLoader, module_name: str
) -> Optional[str]:
    try:
        return loader.get_source(module_name)
    except AttributeError:
        return None  # .get_source() does not supported
    except ImportError:
        return None


def __source_from_file(file_name: str) -> Optional[str]:
    try:
        with open(file_name) as f:
            return f.read()
    except OSError:
        return None


def __get_source(
    file_name: Optional[str] = None,
    loader: Optional[InspectLoader] = None,
    module_name: Optional[str] = None,
) -> Optional[str]:
    src: Optional[str] = None
    if loader and module_name:
        src = __source_from_loader(loader, module_name)
    if not src and file_name:
        src = __source_from_file(file_name)
    return src


def __get_lines(
    line_no: int,
    context_lines: int,
    file_name: Optional[str] = None,
    loader: Optional[InspectLoader] = None,
    module_name: Optional[str] = None,
) -> Optional[SourceInfo]:
    src = __get_source(
        file_name=file_name, loader=loader, module_name=module_name
    )
    if not src:
        return None  # Unable to get the source
    lines = src.splitlines()  # @todo: Implement sliding line iterator
    first_line = max(1, line_no - context_lines)
    return SourceInfo(
        file_name=file_name or module_name or "",
        first_line=first_line,
        current_line=line_no,
        lines=lines[first_line - 1 : line_no + context_lines],
    )



# ---------------------------------------------------------------------
# Gufo Err: Types
# ---------------------------------------------------------------------
# Copyright (C) 2022, Gufo Labs
# ---------------------------------------------------------------------
"""
Public API data types.
"""

# Python modules
from typing import Any, Optional, List, Dict
from dataclasses import dataclass
from uuid import UUID
import datetime


@dataclass
class SourceInfo(object):
    """
    Source context for frame.

    Args:
        file_name: Normalized file name.
        first_line: first line of source context.
        current_line: current execution line.
        lines: List of lines, starting from `first_line`
    """

    file_name: str
    first_line: int
    current_line: int
    lines: List[str]


@dataclass
class FrameInfo(object):
    """
    Execution frame.

    Args:
        name: Current callable name.
        source: Optional `SourceInfo` procedure. May be missed
            on loader problems.
        locals: Dicts of local variables.
        module: Python module name.
    """

    name: str
    source: Optional[SourceInfo]
    locals: Dict[str, Any]
    module: Optional[str] = None


@dataclass
class ErrorInfo(object):
    """
    Current execution frame information.

    Args:
        name: Application or service name.
        version: Application or service version.
        fingerprint: Error fingerprint.
        stack: List of `FrameInfo`. Current execution frame is first.
        exception: Exception instance, if caught.
        timestamp: Error timestamp.
    """

    name: str
    version: str
    fingerprint: UUID
    stack: List[FrameInfo]
    exception: BaseException
    timestamp: Optional[datetime.datetime] = None



# ---------------------------------------------------------------------
# Gufo Err: Serialize/deserialize
# ---------------------------------------------------------------------
# Copyright (C) 2022, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
from typing import Dict, Any, Union, Tuple
import json
import uuid
import datetime

# Gufo Labs modules
from .types import ErrorInfo, FrameInfo, SourceInfo

CODEC_TYPE = "errorinfo"
CURRENT_VERSION = "1.0"


def to_dict(info: ErrorInfo) -> Dict[str, Any]:
    """
    Serialize ErrorInfo to a dict of primitive types.

    Args:
        info: ErrorInfo instance.

    Returns:
        Dict of primitive types (str, int, float).
    """

    def q_x_class(e: BaseException) -> str:
        """
        Get exception class.

        Args:
            e: Exception instance

        Returns:
            Serialized exception class name
        """
        mod = e.__class__.__module__
        ncls = e.__class__.__name__
        if mod == "builtins":
            return ncls
        return f"{mod}.{ncls}"

    def q_var(x: Any) -> Union[str, int, float]:
        """
        Convert variable to the JSON-encodable form.

        Args:
            x: Exception argument

        Returns:
            JSON-serializeable form of argument
        """
        if isinstance(x, (int, float, str)):
            return x
        return str(x)

    def q_frame_info(fi: FrameInfo) -> Dict[str, Any]:
        """
        Convert FrameInfo into JSON-serializeable form.

        Args:
            fi: FrameInfo instance

        Returns:
            Serialized dict
        """
        r = {
            "name": fi.name,
            "module": fi.module,
            "locals": {x: q_var(y) for x, y in fi.locals.items()},
        }
        if fi.source:
            r["source"] = q_source(fi.source)
        return r

    def q_source(si: SourceInfo) -> Dict[str, Any]:
        """
        Convert SourceInfo into JSON-serializeable form.

        Args:
            si: SourceInfo instance

        Returns:
            Serialized dict
        """
        return {
            "file_name": si.file_name,
            "first_line": si.first_line,
            "current_line": si.current_line,
            "lines": si.lines,
        }

    def q_exception(e: BaseException) -> Dict[str, Any]:
        """
        Convery exception into JSON-serializeable form.

        Args:
            e: BaseException instance

        Returns:
            Serialized dict
        """
        return {
            "class": q_x_class(e),
            "args": [q_var(x) for x in e.args],
        }

    r = {
        "$type": CODEC_TYPE,
        "$version": CURRENT_VERSION,
        "name": info.name,
        "version": info.version,
        "fingerprint": str(info.fingerprint),
        "exception": q_exception(info.exception),
        "stack": [q_frame_info(x) for x in info.stack],
    }
    if info.timestamp:
        r["timestamp"] = info.timestamp.isoformat()
    # @todo: stack
    return r


def to_json(info: ErrorInfo) -> str:
    """
    Serialize ErrorInfo to JSON string.

    Args:
        info: ErrorInfo instance.

    Returns:
        json-encoded string.
    """
    return json.dumps(to_dict(info))


def from_dict(data: Dict[str, Any]) -> ErrorInfo:
    """
    Deserealize Dict to ErrorInfo.

    Args:
        data: Result of to_dict

    Returns:
        ErrorInfo instance
    """

    def get(d: Dict[str, Any], name: str) -> Any:
        """
        Get key from dict or raise ValueError if not found.

        Args:
            d: Data dictionary
            name: Key name

        Returns:
            Value
        """
        x = d.get(name, None)
        if x is None:
            raise ValueError(f"{name} is required")
        return x

    def get_fi(d: Dict[str, Any]) -> FrameInfo:
        if d.get("source"):
            source = get_si(d["source"])
        else:
            source = None
        return FrameInfo(
            name=get(d, "name"),
            module=get(d, "module"),
            locals=get(d, "locals"),
            source=source,
        )

    def get_si(d: Dict[str, Any]) -> SourceInfo:
        return SourceInfo(
            file_name=get(d, "file_name"),
            first_line=get(d, "first_line"),
            current_line=get(d, "current_line"),
            lines=get(d, "lines"),
        )

    # Check incoming data is dict
    if not isinstance(data, dict):
        raise ValueError("dict required")
    # Check data has proper type signature
    ci_type = get(data, "$type")
    if ci_type != CODEC_TYPE:
        raise ValueError("Invalid $type")
    # Check version
    ci_version = get(data, "$version")
    if ci_version != CURRENT_VERSION:
        raise ValueError("Unknown $version")
    # Process timestamp
    src_ts = data.get("timestamp")
    if src_ts:
        ts = datetime.datetime.fromisoformat(src_ts)
    else:
        ts = None
    # Exception
    exc = get(data, "exception")
    # Stack
    stack = [get_fi(x) for x in get(data, "stack")]
    # Set exception stub
    return ErrorInfo(
        name=get(data, "name"),
        version=get(data, "version"),
        fingerprint=uuid.UUID(get(data, "fingerprint")),
        timestamp=ts,
        stack=stack,
        exception=ExceptionStub(kls=exc["class"], args=exc["args"]),
    )


def from_json(data: str) -> ErrorInfo:
    """
    Deserialize ErrorInfo from JSON string.

    Args:
        data: JSON string

    Returns:
        ErrorInfo instance
    """
    return from_dict(json.loads(data))


class ExceptionStub(Exception):
    """
    Stub to deserialized exceptions.

    Args:
        kls: Exception class name
        args: Exception arguments
    """

    def __init__(self, kls: str, args: Tuple[Any, ...]) -> None:
        self.kls = kls
        self.args = args



# ---------------------------------------------------------------------
# Gufo Err
# ---------------------------------------------------------------------
# Copyright (C) 2022, Gufo Labs
# ---------------------------------------------------------------------
"""
Human-readable error reporting.

Attributes:
    __version__: Current version.
"""

# Gufo Labs modules
from .types import ErrorInfo, FrameInfo, SourceInfo  # noqa
from .frame import iter_frames, exc_traceback  # noqa
from .logger import logger  # noqa
from .err import Err, err  # noqa
from .abc.failfast import BaseFailFast  # noqa
from .abc.middleware import BaseMiddleware  # noqa

__version__: str = "0.2.0"



# ---------------------------------------------------------------------
# Gufo Err: BaseResponse class
# ---------------------------------------------------------------------
# Copyright (C) 2022, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
from abc import ABC, abstractmethod

# GufoLabs modules
from ..types import ErrorInfo


class BaseMiddleware(ABC):
    """
    Abstract base type for error processing middleware.
    Middleware must implement `process` method.
    """

    @abstractmethod
    def process(self, info: ErrorInfo) -> None:
        """
        Process the error.

        Args:
            info: ErrorInfo instance with detailed error information.
        """
        ...



# ---------------------------------------------------------------------
# Gufo Err: NeverFailFast
# ---------------------------------------------------------------------
# Copyright (C) 2022, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
from typing import Type
from types import TracebackType

# Gufo Labs modules
from ..abc.failfast import BaseFailFast


class NeverFailFast(BaseFailFast):
    """
    Never fail-fast. Always returns False,
    so never inflicts fail-fast.
    """

    def must_die(
        self,
        t: Type[BaseException],
        v: BaseException,
        tb: TracebackType,
    ) -> bool:
        return False



# ---------------------------------------------------------------------
# Gufo Err: BaseFailFast method
# ---------------------------------------------------------------------
# Copyright (C) 2022, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
from abc import ABC, abstractmethod
from typing import Type
from types import TracebackType


class BaseFailFast(ABC):
    """
    Abstract base type for fail-fast behavior. Fail-fast classes
    must implement `must_die` method. When fail-fast check
    decides the error is unrecoverable, it must return `True` value.
    """

    @abstractmethod
    def must_die(
        self,
        t: Type[BaseException],
        v: BaseException,
        tb: TracebackType,
    ) -> bool:
        """
        Fail-fast check. Must be overriden in subclasses.

        Args:
            t: Exception type. Same as `sys.exc_info()[0]`.
            v: Exception value. Same as `sys.exc_info()[1]`.
            tb: Traceback. Same as `sys.exc_info()[2]`.

        Returns:
            * `True`, if the error is not recoverable and the process
            must be terminated ASAP.
            * `False` to pass to the next check.
        """
        ...
